using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;

namespace howto_diagonal_images3
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private int ImgWidth, ImgHeight;
        private float CellWidth, CellHeight, Angle, DividerWidth;
        private Matrix Transform = null, InverseTransform = null;
        private List<Cell> Cells = null;
        private bool DocumentModified = false;

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void mnuFileSaveAs_Click(object sender, EventArgs e)
        {
            SaveResult();
        }
        private void SaveResult()
        {
            if (sfdResult.ShowDialog() == DialogResult.OK)
            {
                using (Bitmap bm = new Bitmap(ImgWidth, ImgHeight))
                {
                    using (Graphics gr = Graphics.FromImage(bm))
                    {
                        DrawCells(gr);
                    }
                    SaveImage(bm, sfdResult.FileName);
                }
                DocumentModified = false;
            }
        }

        private void btnCreate_Click(object sender, EventArgs e)
        {
            if (!DocumentIsSafe()) return;

            try
            {
                // Save the parameters.
                ImgWidth = int.Parse(txtWidth.Text);
                ImgHeight = int.Parse(txtHeight.Text);
                CellWidth = int.Parse(txtCellWidth.Text);
                CellHeight = int.Parse(txtCellHeight.Text);
                Angle = float.Parse(txtAngle.Text);
                DividerWidth = float.Parse(txtDividerWidth.Text);

                Transform = new Matrix();
                Transform.Rotate(Angle);
                InverseTransform = new Matrix();
                InverseTransform.Rotate(-Angle);

                // Make the cells.
                MakeCells();

                // Show the result.
                picCanvas.ClientSize = new Size(ImgWidth, ImgHeight);
                int margin = picCanvas.Left;
                int client_right = margin + Math.Max(
                    picCanvas.Right, btnCreate.Right);
                int client_bottom = margin + picCanvas.Bottom;
                this.ClientSize = new Size(client_right, client_bottom);
                picCanvas.Visible = true;
                mnuFileSaveAs.Enabled = true;

                picCanvas.Refresh();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        // Make the cells.
        private void MakeCells()
        {
            // Rotate the image's corners by -Angle degrees.
            PointF[] points =
            {
                new PointF(0, 0),
                new PointF(0, ImgHeight),
                new PointF(ImgWidth, ImgHeight),
                new PointF(ImgWidth, 0),
            };
            InverseTransform.TransformPoints(points);

            // Get the rotated image's bounds.
            float xmin = points[0].X;
            float ymin = points[0].Y;
            float xmax = xmin;
            float ymax = ymin;
            for (int i = 1; i < points.Length; i++)
            {
                if (xmin > points[i].X) xmin = points[i].X;
                if (xmax < points[i].X) xmax = points[i].X;
                if (ymin > points[i].Y) ymin = points[i].Y;
                if (ymax < points[i].Y) ymax = points[i].Y;
            }

            // Calculate the minimum and maximum rows
            // and columns that might be needed.
            int min_row = (int)(ymin / CellHeight) - 1;
            int max_row = (int)(ymax / CellHeight) + 1;
            int min_col = (int)(xmin / CellWidth) - 1;
            int max_col = (int)(xmax / CellWidth) + 1;

            // Make a GraphicsPath representing the rotated image bounds.
            GraphicsPath image_path = new GraphicsPath();
            image_path.AddPolygon(points);

            // Make a Graphics Object for use in IsEmpty.
            Graphics gr = CreateGraphics();

            // Loop over the possible rows and columns
            // and see which are actually needed.
            Cells = new List<Cell>();
            for (int row = min_row; row <= max_row; row++)
            {
                for (int col = min_col; col <= max_col; col++)
                {
                    // See if this cell's rectangle intersects
                    // the image's rotated bounds.
                    Region rgn = new Region(image_path);
                    float x = col * CellWidth;
                    float y = row * CellHeight;
                    if (Math.Abs(col % 2) == 1) y += CellHeight / 2f;
                    RectangleF cell_rect = new RectangleF(
                        x, y, CellWidth, CellHeight);
                    rgn.Intersect(cell_rect);
                    if (!rgn.IsEmpty(gr))
                    {
                        // Save this cell.
                        Cells.Add(new Cell(cell_rect));
                    }
                }
            }
            Console.WriteLine("# Cells: " + Cells.Count.ToString());
        }

        // Return true if it is safe to discard the current document.
        private bool DocumentIsSafe()
        {
            if (!DocumentModified) return true;
            DialogResult result = MessageBox.Show(
                "Do you want to save your changes?",
                "Save Changes?", MessageBoxButtons.YesNoCancel);

            // The user wants to discard changes.
            if (result == DialogResult.No) return true;

            // The user wants to cancel.
            if (result == DialogResult.Cancel) return false;

            // Try to save the result.
            SaveResult();
            return !DocumentModified;
        }

        private void lblColor_Click(object sender, EventArgs e)
        {
            cdColor.Color = lblColor.BackColor;
            if (cdColor.ShowDialog() == DialogResult.OK)
            {
                lblColor.BackColor = cdColor.Color;
                picCanvas.Refresh();
            }
        }

        // Draw the picture.
        private void picCanvas_Paint(object sender, PaintEventArgs e)
        {
            DrawCells(e.Graphics);
        }
        private void DrawCells(Graphics gr)
        {
            gr.SmoothingMode = SmoothingMode.AntiAlias;
            gr.InterpolationMode = InterpolationMode.High;
            gr.Clear(lblBgColor.BackColor);

            gr.Transform = Transform;
            using (Pen pen = new Pen(lblColor.BackColor, DividerWidth))
            {
                foreach (Cell cell in Cells)
                    cell.Draw(gr, pen, CellWidth, CellHeight);
            }
        }

        // Save the file with the appropriate format.
        public void SaveImage(Image image, string filename)
        {
            string extension = Path.GetExtension(filename);
            switch (extension.ToLower())
            {
                case ".bmp":
                    image.Save(filename, ImageFormat.Bmp);
                    break;
                case ".exif":
                    image.Save(filename, ImageFormat.Exif);
                    break;
                case ".gif":
                    image.Save(filename, ImageFormat.Gif);
                    break;
                case ".jpg":
                case ".jpeg":
                    image.Save(filename, ImageFormat.Jpeg);
                    break;
                case ".png":
                    image.Save(filename, ImageFormat.Png);
                    break;
                case ".tif":
                case ".tiff":
                    image.Save(filename, ImageFormat.Tiff);
                    break;
                default:
                    throw new NotSupportedException(
                        "Unknown file extension " + extension);
            }
        }

        // If there are unsaved changes, ask the user if we should save them.
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            e.Cancel = !DocumentIsSafe();
        }

        // Place a picture in this cell.
        private void picCanvas_MouseClick(object sender, MouseEventArgs e)
        {
            if (ofdCellPicture.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    // Find the clicked cell.
                    // Inverse transform the clicked point.
                    PointF[] points = { e.Location };
                    InverseTransform.TransformPoints(points);

                    // See which cell contains the inverted point.
                    foreach (Cell cell in Cells)
                    {
                        if (cell.ContainsPoint(points[0]))
                        {
                            cell.Picture = new Bitmap(ofdCellPicture.FileName);
                            DocumentModified = true;
                            break;
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }

                picCanvas.Refresh();
            }
        }

        private void txtDividerWidth_TextChanged(object sender, EventArgs e)
        {
            float width;
            if (float.TryParse(txtDividerWidth.Text, out width))
            {
                DividerWidth = width;
                picCanvas.Refresh();
            }
        }

        private void lblBgColor_Click(object sender, EventArgs e)
        {
            cdColor.Color = lblBgColor.BackColor;
            if (cdColor.ShowDialog() == DialogResult.OK)
            {
                lblBgColor.BackColor = cdColor.Color;
                picCanvas.Refresh();
            }
        }
    }
}
